import dask.array as da
import rasterio
from dask import is_dask_collection
from dask.base import tokenize
from rasterio.windows import Window


def read_raster(path, band=None, block_size=1):
    """Read all or some bands from raster

    Arguments:
        path {string} -- path to raster file

    Keyword Arguments:
        band {int, iterable(int)} -- band number or iterable of bands.
            When passing None, it reads all bands (default: {None})
        block_size {int} -- block size multiplier (default: {1})

    Returns:
        dask.array.Array -- a Dask array
    """

    if isinstance(band, int):
        return read_raster_band(path, band=band, block_size=block_size)
    else:
        if band is None:
            bands = range(1, get_band_count(path) + 1)
        else:
            bands = list(band)
        return da.stack([
            read_raster_band(path, band=band, block_size=block_size)
            for band in bands
        ])


def read_raster_band(path, band=1, block_size=1):
    """Read a raster band and return a Dask array

    Arguments:
        path {string} -- path to the raster file

    Keyword Arguments:
        band {int} -- number of band to read (default: {1})
        block_size {int} -- block size multiplier (default: {1})

    """

    def read_window(raster_path, window, band):
        with rasterio.open(raster_path) as src:
            return src.read(band, window=window)

    def resize_window(window, block_size):
        return Window(
            col_off=window.col_off * block_size,
            row_off=window.row_off * block_size,
            width=window.width * block_size,
            height=window.height * block_size)

    def block_windows(dataset, band, block_size):
        return [(pos, resize_window(win, block_size))
                for pos, win in dataset.block_windows(band)]

    with rasterio.open(path) as src:
        h, w = src.block_shapes[band - 1]
        chunks = (h * block_size, w * block_size)
        name = 'raster-{}'.format(tokenize(path, band, chunks))
        dtype = src.dtypes[band - 1]
        shape = src.shape
        blocks = block_windows(src, band, block_size)

    dsk = {(name, i, j): (read_window, path, window, band)
           for (i, j), window in blocks}

    return da.Array(dsk, name, chunks, dtype, shape)


def get_band_count(raster_path):
    """Read raster band count"""
    with rasterio.open(raster_path) as src:
        return src.count



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['dask_rasterio']

package_data = \
{'': ['*']}

install_requires = \
['dask>=0.17.5,<0.18.0', 'rasterio>=1.0b1,<2.0', 'toolz>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'dask-rasterio',
    'version': '0.2.1',
    'description': 'Read and write rasters in parallel using Rasterio and Dask',
    'long_description': "# dask-rasterio\n\n[![Build Status](https://travis-ci.org/dymaxionlabs/dask-rasterio.svg?branch=master)](https://travis-ci.org/dymaxionlabs/dask-rasterio)\n[![codecov](https://codecov.io/gh/dymaxionlabs/dask-rasterio/branch/master/graph/badge.svg)](https://codecov.io/gh/dymaxionlabs/dask-rasterio) [![Join the chat at https://gitter.im/dymaxionlabs/dask-rasterio](https://badges.gitter.im/dymaxionlabs/dask-rasterio.svg)](https://gitter.im/dymaxionlabs/dask-rasterio?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n\n`dask-rasterio` provides some methods for reading and writing rasters in\nparallel using [Rasterio](https://github.com/mapbox/rasterio) and\n[Dask](https://dask.pydata.org) arrays.\n\n\n## Usage\n\n#### Read a multiband raster\n\n```python\n>>> from dask_rasterio import read_raster\n\n>>> array = read_raster('tests/data/RGB.byte.tif')\n>>> array\ndask.array<stack, shape=(3, 718, 791), dtype=uint8, chunksize=(1, 3, 791)>\n\n>>> array.mean()\ndask.array<mean_agg-aggregate, shape=(), dtype=float64, chunksize=()>\n>>> array.mean().compute()\n40.858976977533935\n```\n\n#### Read a single band from a raster\n\n```python\n>>> from dask_rasterio import read_raster\n\n>>> array = read_raster('tests/data/RGB.byte.tif', band=3)\n>>> array\ndask.array<raster, shape=(718, 791), dtype=uint8, chunksize=(3, 791)>\n```\n\n#### Write a singleband or multiband raster\n\n```python\n>>> from dask_rasterio import read_raster, write_raster\n\n>>> array = read_raster('tests/data/RGB.byte.tif')\n\n>>> new_array = array & (array > 100)\n>>> new_array\ndask.array<and_, shape=(3, 718, 791), dtype=uint8, chunksize=(1, 3, 791)>\n\n>>> prof = ... # reuse profile from tests/data/RGB.byte.tif...\n>>> write_raster('processed_image.tif', new_array, **prof)\n```\n\n#### Chunk size\n\nBoth `read_raster` and `write_raster` accept a `block_size` argument that\nacts as a multiplier to the block size of rasters. The default value is 1,\nwhich means the dask array chunk size will be the same as the block size of\nthe raster file. You will have to adjust this value depending on the\nspecification of your machine (how much memory do you have, and the block\nsize of the raster).\n\n\n## Install\n\nInstall with pip:\n\n```\npip install dask-rasterio\n```\n\n## Development\n\nThis project is managed by [Poetry](https://github.com/sdispater/poetry).  If\nyou do not have it installed, please refer to \n[Poetry instructions](https://github.com/sdispater/poetry#installation).\n\nNow, clone the repository and run `poetry install`.  This will create a virtual\nenvironment and install all required packages there.\n\nRun `poetry run pytest` to run all tests.\n\nRun `poetry build` to build package on `dist/`.\n\n\n## Issue tracker\n\nPlease report any bugs and enhancement ideas using the GitHub issue tracker:\n\n  https://github.com/dymaxionlabs/dask-rasterio/issues\n\nFeel free to also ask questions on our\n[Gitter channel](https://gitter.im/dymaxionlabs/dask-rasterio), or by email.\n\n\n## Help wanted\n\nAny help in testing, development, documentation and other tasks is highly\nappreciated and useful to the project.\n\nFor more details, see the file [CONTRIBUTING.md](CONTRIBUTING.md).\n\n\n## License\n\nSource code is released under a BSD-2 license.  Please refer to\n[LICENSE.md](LICENSE.md) for more information.\n",
    'author': 'Damián Silvani',
    'author_email': 'munshkr@gmail.com',
    'url': 'https://github.com/dymaxionlabs/dask-rasterio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3,<4',
}


setup(**setup_kwargs)



import dask.array as da
import rasterio
from dask import is_dask_collection
from dask.base import tokenize
from rasterio.windows import Window


def write_raster(path, array, **kwargs):
    """Write a dask array to a raster file

    If array is 2d, write array on band 1.
    If array is 3d, write data on each band

    Arguments:
        path {string} -- path of raster to write
        array {dask.array.Array} -- band array
        kwargs {dict} -- keyword arguments to delegate to rasterio.open

    Examples:
        # Write a single band raster
        >> red_band = read_raster_band("test.tif", band=1)
        >> write_raster("new.tif", red_band)

        # Write a multiband raster
        >> img = read_raster("test.tif")
        >> new_img = process(img)
        >> write_raster("new.tif", new_img)

    """
    if len(array.shape) != 2 and len(array.shape) != 3:
        raise TypeError('invalid shape (must be either 2d or 3d)')

    if is_dask_collection(array):
        with RasterioDataset(path, 'w', **kwargs) as dst:
            da.store(array, dst, lock=True)
    else:
        with rasterio.open(path, 'w', **kwargs) as dst:
            if len(array.shape) == 2:
                dst.write(array, 1)
            else:
                dst.write(array)


class RasterioDataset:
    """Rasterio wrapper to allow dask.array.store to do window saving.

    Example:
        >> rows = cols = 21696
        >> a = da.ones((4, rows, cols), dtype=np.float64, chunks=(1, 4096, 4096) )
        >> a = a * np.array([255., 255., 255., 255.])[:, None, None]
        >> a = a.astype(np.uint8)
        >> with RasterioDataset('test.tif', 'w', driver='GTiff', width=cols, height=rows, count=4, dtype=np.uint8) as r_file:
        ..    da.store(a, r_file, lock=True)
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.dataset = None

    def __setitem__(self, key, item):
        """Put the data chunk in the image"""
        if len(key) == 3:
            index_range, y, x = key
            indexes = list(
                range(index_range.start + 1, index_range.stop + 1,
                      index_range.step or 1))
        else:
            indexes = 1
            y, x = key

        chy_off = y.start
        chy = y.stop - y.start
        chx_off = x.start
        chx = x.stop - x.start

        self.dataset.write(
            item, window=Window(chx_off, chy_off, chx, chy), indexes=indexes)

    def __enter__(self):
        """Enter method"""
        self.dataset = rasterio.open(*self.args, **self.kwargs)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit method"""
        self.dataset.close()
